@@ -11,12 +11,10 @@
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
-from mercurial import error
-
-from tortoisehg.util import paths, hglib
+from tortoisehg.util import hglib
 
 from tortoisehg.hgqt.i18n import _
-from tortoisehg.hgqt import qtlib, fileview, status, thgrepo, filectxactions
+from tortoisehg.hgqt import qtlib, fileview, status, filectxactions
 from tortoisehg.hgqt import revpanel
 from tortoisehg.hgqt.manifestmodel import ManifestModel
 
@@ -28,13 +26,13 @@ class ManifestDialog(QMainWindow):
     finished = pyqtSignal(int)
     linkActivated = pyqtSignal(QString)
 
-    def __init__(self, repo, rev=None, parent=None):
+    def __init__(self, repoagent, rev=None, parent=None):
         QMainWindow.__init__(self, parent)
-        self._repo = repo
+        self._repoagent = repoagent
         self.setWindowIcon(qtlib.geticon('hg-annotate'))
         self.resize(400, 300)
 
-        self._manifest_widget = ManifestWidget(repo, rev)
+        self._manifest_widget = ManifestWidget(repoagent, rev)
         self._manifest_widget.revChanged.connect(self._updatewindowtitle)
         self._manifest_widget.pathChanged.connect(self._updatewindowtitle)
         self._manifest_widget.grepRequested.connect(self._openSearchWidget)
@@ -46,6 +44,9 @@ class ManifestDialog(QMainWindow):
         self._manifest_widget.showMessage.connect(self.statusBar().showMessage)
         self._manifest_widget.linkActivated.connect(self._linkHandler)
 
+        self._dialogs = qtlib.DialogKeeper(
+            lambda self, dlgmeth: dlgmeth(self), parent=self)
+
         self._readsettings()
         self._updatewindowtitle()
 
@@ -69,6 +70,9 @@ class ManifestDialog(QMainWindow):
         s.setValue('manifest/geom', self.saveGeometry())
         self._manifest_widget.saveSettings(s, 'manifest')
 
+    def setRev(self, rev):
+        self._manifest_widget.setRev(rev)
+
     def setSource(self, path, rev, line=None):
         self._manifest_widget.setSource(path, rev, line)
 
@@ -80,18 +84,27 @@ class ManifestDialog(QMainWindow):
         """Set if search is case insensitive"""
         self._manifest_widget._fileview.searchbar.setCaseInsensitive(ignorecase)
 
+    def setFileViewMode(self, mode):
+        self._manifest_widget.setFileViewMode(mode)
+
     @pyqtSlot(unicode, dict)
     def _openSearchWidget(self, pattern, opts):
+        dlg = self._dialogs.open(ManifestDialog._createSearchDialog)
         opts = dict((str(k), str(v)) for k, v in opts.iteritems())
-        from tortoisehg.hgqt import run
-        run.grep(self._repo.ui, hglib.fromunicode(pattern), **opts)
+        dlg.setSearch(pattern, **opts)
+        dlg.runSearch()
+
+    def _createSearchDialog(self):
+        from tortoisehg.hgqt import grep
+        return grep.SearchDialog(self._repoagent, [])
 
     @pyqtSlot(QString)
     def _linkHandler(self, link):
         ulink = unicode(link)
         if ulink.startswith('cset:'):
+            repo = self._repoagent.rawRepo()
             changeid = hglib.fromunicode(ulink[len('cset:'):])
-            rev = self._repo[changeid].rev()
+            rev = repo[changeid].rev()
             self._manifest_widget.setRev(rev)
         else:
             self.linkActivated.emit(link)
@@ -118,12 +131,17 @@ class ManifestWidget(QWidget, qtlib.TaskWidget):
     revsetFilterRequested = pyqtSignal(QString)
     """Ask the repowidget to change its revset filter"""
 
+    runCustomCommandRequested = pyqtSignal(str, list)
+    """Emitted when selects a custom tool on the context menu"""
+
     def canswitch(self):
         return False
 
-    def __init__(self, repo, rev=None, parent=None):
+    def __init__(self, repoagent, rev=None, parent=None):
         super(ManifestWidget, self).__init__(parent)
-        self._repo = repo
+        self._repoagent = repoagent
+        # TODO: replace by repoagent if setRepo(bundlerepo) can be removed
+        self._repo = repoagent.rawRepo()
         self._rev = rev
         self._selectedrev = rev
 
@@ -163,7 +181,7 @@ class ManifestWidget(QWidget, qtlib.TaskWidget):
         self.revpanel = revpanel.RevPanelWidget(self._repo)
         self.revpanel.linkActivated.connect(self.linkActivated)
         vbox.addWidget(self.revpanel, 0)
-        self._fileview = fileview.HgFileView(self._repo, self)
+        self._fileview = fileview.HgFileView(self._repoagent, self)
         vbox.addWidget(self._fileview, 0)
         w = QWidget()
         w.setLayout(vbox)
@@ -207,6 +225,8 @@ class ManifestWidget(QWidget, qtlib.TaskWidget):
                                                           rev=self._rev)
         self._fileactions.linkActivated.connect(self.linkActivated)
         self._fileactions.filterRequested.connect(self.revsetFilterRequested)
+        self._fileactions.runCustomCommandRequested.connect(
+            self.runCustomCommandRequested)
         self.addActions(self._fileactions.actions())
 
     def showEvent(self, event):
@@ -392,6 +412,9 @@ class ManifestWidget(QWidget, qtlib.TaskWidget):
     def _updatecontent(self):
         self.displayFile()
 
+    def setFileViewMode(self, mode):
+        self._fileview.setMode(mode)
+
     @pyqtSlot()
     def _emitPathChanged(self):
         self.pathChanged.emit(self.path)
@@ -401,48 +424,6 @@ def connectsearchbar(manifestwidget, searchbar):
     searchbar.conditionChanged.connect(manifestwidget.highlightText)
     searchbar.searchRequested.connect(manifestwidget.find)
 
-def run(ui, *pats, **opts):
-    repo = opts.get('repo') or thgrepo.repository(ui, paths.find_root())
-    try:
-        # ManifestWidget expects integer revision
-        rev = repo[opts.get('rev')].rev()
-    except error.RepoLookupError, e:
-        qtlib.ErrorMsgBox(_('Failed to open Manifest dialog'),
-                          hglib.tounicode(e.message))
-        return
-    try:
-        line = opts.get('line') and int(opts['line']) or None
-    except ValueError:
-        qtlib.ErrorMsgBox(_('Failed to open Manifest dialog'),
-                          _('The specified line number "%s" is invalid.')
-                          % hglib.tounicode(opts.get('line')))
-        return
-
-    dlg = ManifestDialog(repo, rev)
-
-    # set initial state after dialog visible
-    def init():
-        try:
-            if pats:
-                path = hglib.canonpaths(pats)[0]
-            elif 'canonpath' in opts:
-                path = opts['canonpath']
-            else:
-                return
-            dlg.setSource(hglib.tounicode(path), rev, line)
-            if opts.get('pattern'):
-                dlg.setSearchPattern(opts['pattern'])
-            if dlg._manifest_widget._fileview.actionAnnMode.isEnabled():
-                dlg._manifest_widget._fileview.actionAnnMode.trigger()
-            if 'ignorecase' in opts:
-                dlg.setSearchCaseInsensitive(opts['ignorecase'])
-        except IndexError:
-            pass
-        dlg.setSearchPattern(hglib.tounicode(opts.get('pattern')) or '')
-    QTimer.singleShot(0, init)
-
-    return dlg
-
 # In order to let the user seamlessly switch between the filterbox and the treeview
 # we subclas the QLineEdit and QTreeView widgets. We add some keypress related signals
 # will be used by the ManifestWidget to change the focus between these two widgets
