@@ -9,28 +9,23 @@ import os
 import sys
 import atexit
 import shutil
+import shlex
 import stat
 import subprocess
 import tempfile
 import re
 import weakref
 
+from mercurial.i18n import _ as hggettext
 from mercurial import commands, extensions, error, util
 
-from tortoisehg.util import hglib, paths
+from tortoisehg.util import hglib, paths, editor, terminal
 from tortoisehg.hgqt.i18n import _
 from hgext.color import _styles
 
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 
-if PYQT_VERSION_STR.split('.') < ['4', '7'] or \
-   QT_VERSION_STR.split('.') < ['4', '6']:
-    sys.stderr.write('TortoiseHg requires Qt 4.6 and PyQt 4.7\n')
-    sys.stderr.write('You have Qt %s and PyQt %s\n' %
-                     (QT_VERSION_STR, PYQT_VERSION_STR))
-    sys.exit()
-
 try:
     import win32con
     openflags = win32con.CREATE_NO_WINDOW
@@ -58,7 +53,7 @@ def gettempdir():
 def openhelpcontents(url):
     'Open online help, use local CHM file if available'
     if not url.startswith('http'):
-        fullurl = 'http://tortoisehg.org/manual/2.4/' + url
+        fullurl = 'http://tortoisehg.org/manual/2.7/' + url
         # Use local CHM file if it can be found
         if os.name == 'nt' and paths.bin_path:
             chm = os.path.join(paths.bin_path, 'doc', 'TortoiseHg.chm')
@@ -92,8 +87,9 @@ def openfiles(repo, files, parent=None):
     for filename in files:
         openlocalurl(repo.wjoin(filename))
 
-def editfiles(repo, files, lineno=None, search=None, parent=None, editor=None):
+def editfiles(repo, files, lineno=None, search=None, parent=None):
     if len(files) == 1:
+        # if editing a single file, open in cwd context of that file
         filename = files[0].strip()
         if not filename:
             return
@@ -102,20 +98,55 @@ def editfiles(repo, files, lineno=None, search=None, parent=None, editor=None):
         cwd = os.path.dirname(path)
         files = [os.path.basename(path)]
     else:
+        # else edit in cwd context of repo root
         cwd = repo.root
+
+    toolpath, args, argsln, argssearch = editor.detecteditor(repo, files)
+    if os.path.basename(toolpath) in ('vi', 'vim', 'hgeditor'):
+        res = QMessageBox.critical(parent,
+                    _('No visual editor configured'),
+                    _('Please configure a visual editor.'))
+        from tortoisehg.hgqt.settings import SettingsDialog
+        dlg = SettingsDialog(False, focus='tortoisehg.editor')
+        dlg.exec_()
+        return
+
     files = [util.shellquote(util.localpath(f)) for f in files]
-    editor = repo.ui.config('tortoisehg', 'editor')
     assert len(files) == 1 or lineno == None
-    if not editor:
-        editor = repo.ui.config('tortoisehg', 'editor')
-    if editor:
+
+    cmdline = None
+    if search:
+        assert lineno is not None
+        if argssearch:
+            cmdline = ' '.join([toolpath, argssearch])
+            cmdline = cmdline.replace('$LINENUM', str(lineno))
+            cmdline = cmdline.replace('$SEARCH', search)
+        elif argsln:
+            cmdline = ' '.join([toolpath, argsln])
+            cmdline = cmdline.replace('$LINENUM', str(lineno))
+        elif args:
+            cmdline = ' '.join([toolpath, args])
+    elif lineno:
+        if argsln:
+            cmdline = ' '.join([toolpath, argsln])
+            cmdline = cmdline.replace('$LINENUM', str(lineno))
+        elif args:
+            cmdline = ' '.join([toolpath, args])
+    else:
+        if args:
+            cmdline = ' '.join([toolpath, args])
+
+    if cmdline is None:
+        # editor was not specified by editor-tools configuration, fall
+        # back to older tortoisehg.editor OpenAtLine parsing
+        cmdline = ' '.join([toolpath] + files) # default
         try:
             regexp = re.compile('\[([^\]]*)\]')
             expanded = []
             pos = 0
-            for m in regexp.finditer(editor):
-                expanded.append(editor[pos:m.start()-1])
-                phrase = editor[m.start()+1:m.end()-1]
+            for m in regexp.finditer(toolpath):
+                expanded.append(toolpath[pos:m.start()-1])
+                phrase = toolpath[m.start()+1:m.end()-1]
                 pos = m.end()+1
                 if '$LINENUM' in phrase:
                     if lineno is None:
@@ -131,50 +162,54 @@ def editfiles(repo, files, lineno=None, search=None, parent=None, editor=None):
                     phrase = phrase.replace('$FILE', files[0])
                     files = []
                 expanded.append(phrase)
-            expanded.append(editor[pos:])
+            expanded.append(toolpath[pos:])
             cmdline = ' '.join(expanded + files)
         except ValueError, e:
             # '[' or ']' not found
-            cmdline = ' '.join([editor] + files)
+            pass
         except TypeError, e:
             # variable expansion failed
-            cmdline = ' '.join([editor] + files)
-    else:
-        editor = os.environ.get('HGEDITOR') or repo.ui.config('ui', 'editor') \
-                 or os.environ.get('EDITOR', 'vi')
-        cmdline = ' '.join([editor] + files)
-    if os.path.basename(editor) in ('vi', 'vim', 'hgeditor'):
-        res = QMessageBox.critical(parent,
-                    _('No visual editor configured'),
-                    _('Please configure a visual editor.'))
-        from tortoisehg.hgqt.settings import SettingsDialog
-        dlg = SettingsDialog(False, focus='tortoisehg.editor')
-        dlg.exec_()
-        return
+            pass
 
-    cmdline = util.quotecommand(cmdline)
     shell = not (len(cwd) >= 2 and cwd[0:2] == r'\\')
     try:
-        subprocess.Popen(cmdline, shell=shell, creationflags=openflags,
-                         stderr=None, stdout=None, stdin=None, cwd=cwd)
+        if '$FILES' in cmdline:
+            cmdline = cmdline.replace('$FILES', ' '.join(files))
+            cmdline = util.quotecommand(cmdline)
+            subprocess.Popen(cmdline, shell=shell, creationflags=openflags,
+                             stderr=None, stdout=None, stdin=None, cwd=cwd)
+        elif '$FILE' in cmdline:
+            for file in files:
+                cmd = cmdline.replace('$FILE', file)
+                cmd = util.quotecommand(cmd)
+                subprocess.Popen(cmd, shell=shell, creationflags=openflags,
+                                 stderr=None, stdout=None, stdin=None, cwd=cwd)
+        else:
+            # assume filenames were expanded already
+            cmdline = util.quotecommand(cmdline)
+            subprocess.Popen(cmdline, shell=shell, creationflags=openflags,
+                             stderr=None, stdout=None, stdin=None, cwd=cwd)
     except (OSError, EnvironmentError), e:
         QMessageBox.warning(parent,
                 _('Editor launch failure'),
                 u'%s : %s' % (hglib.tounicode(cmdline),
                               hglib.tounicode(str(e))))
-    return False
 
 def savefiles(repo, files, rev, parent=None):
     for curfile in files:
         wfile = util.localpath(curfile)
         wfile, ext = os.path.splitext(os.path.basename(wfile))
+        extfilter = [_("All files (*)")]
         if wfile:
             filename = "%s@%d%s" % (wfile, rev, ext)
+            if ext:
+                extfilter.insert(0, "*%s" % ext)
         else:
             filename = "%s@%d" % (ext, rev)
-        result = QFileDialog.getSaveFileName(
-            parent=parent, caption=_("Save file to"),
-            directory=hglib.tounicode(filename))
+
+        result = QFileDialog.getSaveFileName(parent, _("Save file to"),
+                                             hglib.tounicode(filename),
+                                             ";;".join(extfilter))
         if not result:
             continue
         cwd = os.getcwd()
@@ -184,24 +219,43 @@ def savefiles(repo, files, rev, parent=None):
                 commands.cat(repo.ui, repo, curfile, rev=rev,
                              output=hglib.fromunicode(result))
             except (util.Abort, IOError), e:
-                QMessageBox.critical(self, _('Unable to save file'),
+                QMessageBox.critical(parent, _('Unable to save file'),
                                      hglib.tounicode(str(e)))
         finally:
             os.chdir(cwd)
 
-_user_shell = None
-def openshell(root, reponame):
+def openshell(root, reponame, ui=None):
     if not os.path.exists(root):
         WarningMsgBox(
             _('Failed to open path in terminal'),
             _('"%s" is not a valid directory') % hglib.tounicode(root))
         return
-    if _user_shell:
+    shell, args = terminal.detectterminal(ui)
+    if shell:
+        if args:
+            shell = shell + ' ' + util.expandpath(args)
+        # check invalid expression in tortoisehg.shell.  we shouldn't apply
+        # string formatting to untrusted value, but too late to change syntax.
+        try:
+            shell % {'root': '', 'reponame': ''}
+        except (KeyError, TypeError, ValueError):
+            # KeyError: "%(invalid)s", TypeError: "%(root)d", ValueError: "%"
+            ErrorMsgBox(_('Failed to open path in terminal'),
+                        _('Invalid configuration: %s')
+                        % hglib.tounicode(shell))
+            return
+        shellcmd = shell % {'root': root, 'reponame': reponame}
+
         cwd = os.getcwd()
         try:
-            shellcmd = _user_shell % {'reponame': reponame}
+            # Unix: QProcess.startDetached(program) cannot parse single-quoted
+            # parameters built using util.shellquote().
+            # Windows: subprocess.Popen(program, shell=True) cannot spawn
+            # cmd.exe in new window, probably because the initial cmd.exe is
+            # invoked with SW_HIDE.
             os.chdir(root)
-            started = QProcess.startDetached(shellcmd)
+            fullargs = map(hglib.tounicode, shlex.split(shellcmd))
+            started = QProcess.startDetached(fullargs[0], fullargs[1:])
         finally:
             os.chdir(cwd)
         if not started:
@@ -211,17 +265,12 @@ def openshell(root, reponame):
         InfoMsgBox(_('No shell configured'),
                    _('A terminal shell must be configured'))
 
-def configureshell(ui):
-    global _user_shell
-    _user_shell = ui.config('tortoisehg', 'shell')
-    if _user_shell:
-        return
-    if sys.platform == 'darwin':
-        return # Terminal.App does not support open-to-folder
-    elif os.name == 'nt':
-        _user_shell = 'cmd.exe /K title %(reponame)s'
-    else:
-        _user_shell = 'xterm -T "%(reponame)s"'
+
+def isdarktheme(palette=None):
+    """True if white-on-black color scheme is preferable"""
+    if not palette:
+        palette = QApplication.palette()
+    return palette.color(QPalette.Base).black() >= 0x80
 
 # _styles maps from ui labels to effects
 # _effects maps an effect to font style properties.  We define a limited
@@ -255,8 +304,6 @@ thgstylesheet = '* { white-space: pre; font-family: monospace;' \
 tbstylesheet = 'QToolBar { border: 0px }'
 
 def configstyles(ui):
-    configureshell(ui)
-
     # extensions may provide more labels and default effects
     for name, ext in extensions.extensions():
         _styles.update(getattr(ext, 'colortable', {}))
@@ -441,6 +488,14 @@ def descriptionhtmlizer(ui):
 
 _iconcache = {}
 
+if hasattr(QIcon, 'hasThemeIcon'):  # PyQt>=4.7
+    def _findthemeicon(name):
+        if QIcon.hasThemeIcon(name):
+            return QIcon.fromTheme(name)
+else:
+    def _findthemeicon(name):
+        pass
+
 def _findicon(name):
     # TODO: icons should be placed at single location before release
     if os.path.isabs(name):
@@ -466,6 +521,17 @@ _SCALABLE_ICON_PATHS = [(QSize(), 'scalable/actions', '.svg'),
                         (QSize(32, 32), '32x32/actions', '.png'),
                         (QSize(24, 24), '24x24/actions', '.png')]
 
+def getallicons():
+    """Get a sorted, unique list of all available icons"""
+    iconset = set()
+    for size, subdir, sfx in _SCALABLE_ICON_PATHS:
+        path = ':/icons/%s' % (subdir)
+        d = QDir(path)
+        d.setNameFilters(['*%s' % sfx])
+        for iconname in d.entryList():
+            iconset.add(unicode(iconname).rsplit('.', 1)[0])
+    return sorted(iconset)
+
 def _findscalableicon(name):
     """Find icon from qrc by using freedesktop-like icon lookup"""
     o = QIcon()
@@ -485,23 +551,15 @@ def geticon(name):
     This searches for the icon from theme, Qt resource or icons directory,
     named as 'name.(svg|png|ico)'.
     """
-    if QIcon.hasThemeIcon(name):
-        return QIcon.fromTheme(name)
-
     try:
         return _iconcache[name]
     except KeyError:
-        _iconcache[name] = (_findscalableicon(name) or _findicon(name)
+        _iconcache[name] = (_findthemeicon(name)
+                            or _findscalableicon(name)
+                            or _findicon(name)
                             or QIcon(':/icons/fallback.svg'))
         return _iconcache[name]
 
-if sys.platform == 'darwin':
-    # On Mac OS X, we do not want icons on menus
-    def getmenuicon(name):
-        return QIcon()
-else:
-    getmenuicon = geticon
-
 
 def getoverlaidicon(base, overlay):
     """Generate an overlaid icon"""
@@ -524,6 +582,19 @@ def getpixmap(name, width=16, height=16):
     _pixmapcache[key] = pixmap
     return pixmap
 
+def getcheckboxpixmap(state, bgcolor, widget):
+    pix = QPixmap(16,16)
+    painter = QPainter(pix)
+    painter.fillRect(0, 0, 16, 16, bgcolor)
+    option = QStyleOptionButton()
+    style = QApplication.style()
+    option.initFrom(widget)
+    option.rect = style.subElementRect(style.SE_CheckBoxIndicator, option)
+    option.rect.moveTo(1, 1)
+    option.state |= state
+    style.drawPrimitive(style.PE_IndicatorCheckBox, option, painter)
+    return pix
+
 class ThgFont(QObject):
     changed = pyqtSignal(QFont)
     def __init__(self, name):
@@ -550,19 +621,12 @@ _fontcache = {}
 def initfontcache(ui):
     for name in _fontdefaults:
         fname = ui.config('tortoisehg', name, _fontdefaults[name])
-        _fontcache[name] = ThgFont(fname)
+        _fontcache[name] = ThgFont(hglib.tounicode(fname))
 
 def getfont(name):
     assert name in _fontdefaults
     return _fontcache[name]
 
-def gettranslationpath():
-    """Return path to Qt's translation file (.qm)"""
-    if getattr(sys, 'frozen', False):
-        return ':/translations'
-    else:
-        return QLibraryInfo.location(QLibraryInfo.TranslationsPath)
-
 def CommonMsgBox(icon, title, main, text='', buttons=QMessageBox.Ok,
                  labels=[], parent=None, defaultbutton=None):
     msg = QMessageBox(parent)
@@ -669,29 +733,18 @@ class ChoicePrompt(QDialog):
             return self.choices[self.choice_combo.currentIndex()]
         return None
 
-def setup_font_substitutions():
-    QFont.insertSubstitutions('monospace', ['monaco', 'courier new'])
+def allowCaseChangingInput(combo):
+    """Allow case-changing input of known combobox item
 
-def fix_application_font():
-    if os.name != 'nt':
-        return
-    try:
-        import win32gui, win32con
-    except ImportError:
-        return
+    QComboBox performs case-insensitive inline completion by default. It's
+    all right, but sadly it implies case-insensitive check for duplicates,
+    i.e. you can no longer enter "Foo" if the combobox contains "foo".
 
-    # use configurable font like GTK, Mozilla XUL or Eclipse SWT
-    ncm = win32gui.SystemParametersInfo(win32con.SPI_GETNONCLIENTMETRICS)
-    lf = ncm['lfMessageFont']
-    f = QFont(hglib.tounicode(lf.lfFaceName))
-    f.setItalic(lf.lfItalic)
-    if lf.lfWeight != win32con.FW_DONTCARE:
-        weights = [(0, QFont.Light), (400, QFont.Normal), (600, QFont.DemiBold),
-                   (700, QFont.Bold), (800, QFont.Black)]
-        n, w = filter(lambda e: e[0] <= lf.lfWeight, weights)[-1]
-        f.setWeight(w)
-    f.setPixelSize(abs(lf.lfHeight))
-    QApplication.setFont(f, 'QWidget')
+    For details, read QComboBoxPrivate::_q_editingFinished() and matchFlags()
+    of src/gui/widgets/qcombobox.cpp.
+    """
+    assert isinstance(combo, QComboBox) and combo.isEditable()
+    combo.completer().setCaseSensitivity(Qt.CaseSensitive)
 
 class PMButton(QPushButton):
     """Toggle button with plus/minus icon images"""
@@ -710,10 +763,12 @@ class PMButton(QPushButton):
         icon = expanded and self.minus or self.plus
         self.setIcon(icon)
 
-        def clicked():
-            icon = self.is_expanded() and self.plus or self.minus
-            self.setIcon(icon)
-        self.clicked.connect(clicked)
+        self.clicked.connect(self._toggle_icon)
+
+    @pyqtSlot()
+    def _toggle_icon(self):
+        icon = self.is_expanded() and self.plus or self.minus
+        self.setIcon(icon)
 
     def set_expanded(self, state=True):
         icon = state and self.minus or self.plus
@@ -755,7 +810,7 @@ class ExpanderLabel(QWidget):
         self.button.clicked.connect(self.pm_clicked)
         box.addWidget(self.button)
         self.label = ClickableLabel(label, self)
-        self.label.clicked.connect(lambda: self.button.click())
+        self.label.clicked.connect(self.button.click)
         box.addWidget(self.label)
         if not stretch:
             box.addStretch(0)
@@ -843,16 +898,65 @@ class LabeledSeparator(QWidget):
 
         self.setLayout(box)
 
+# Strings and regexes used to convert hashes and subrepo paths into links
+_hashregex = re.compile(r'\b[0-9a-fA-F]{12,}')
+# Currently converting subrepo paths into links only works in English
+_subrepoindicatorpattern = hglib.tounicode(hggettext('(in subrepo %s)') + '\n')
+
+def _linkifyHash(message, subrepo=''):
+    if subrepo:
+        p = 'repo:%s?' % subrepo
+    else:
+        p = 'cset:'
+    replaceexpr = lambda m: '<a href="%s">%s</a>' % (p + m.group(0), m.group(0))
+    return _hashregex.sub(replaceexpr, message)
+
+def _linkifySubrepoRef(message, subrepo, hash=''):
+    if hash:
+        hash = '?' + hash
+    subrepolink = '<a href="repo:%s%s">%s</a>' % (subrepo, hash, subrepo)
+    subrepoindicator = _subrepoindicatorpattern % subrepo
+    linkifiedsubrepoindicator = _subrepoindicatorpattern % subrepolink
+    message = message.replace(subrepoindicator, linkifiedsubrepoindicator)
+    return message
+
+def linkifyMessage(message, subrepo=None):
+    r"""Convert revision id hashes and subrepo paths in messages into links
+
+    >>> linkifyMessage('abort: 0123456789ab!\nhint: foo\n')
+    u'abort: <a href="cset:0123456789ab">0123456789ab</a>!<br>hint: foo<br>'
+    >>> linkifyMessage('abort: foo (in subrepo bar)\n', subrepo='bar')
+    u'abort: foo (in subrepo <a href="repo:bar">bar</a>)<br>'
+    >>> linkifyMessage('abort: 0123456789ab! (in subrepo bar)\nhint: foo\n',
+    ...                subrepo='bar') #doctest: +NORMALIZE_WHITESPACE
+    u'abort: <a href="repo:bar?0123456789ab">0123456789ab</a>!
+    (in subrepo <a href="repo:bar?0123456789ab">bar</a>)<br>hint: foo<br>'
+
+    subrepo name containing regexp backreference, \g:
+
+    >>> linkifyMessage('abort: 0123456789ab! (in subrepo foo\\goo)\n',
+    ...                subrepo='foo\\goo') #doctest: +NORMALIZE_WHITESPACE
+    u'abort: <a href="repo:foo\\goo?0123456789ab">0123456789ab</a>!
+    (in subrepo <a href="repo:foo\\goo?0123456789ab">foo\\goo</a>)<br>'
+    """
+    message = unicode(message)
+    message = _linkifyHash(message, subrepo)
+    if subrepo:
+        hash = ''
+        m = _hashregex.search(message)
+        if m:
+            hash = m.group(0)
+        message = _linkifySubrepoRef(message, subrepo, hash)
+    return message.replace('\n', '<br>')
+
 class InfoBar(QFrame):
     """Non-modal confirmation/alert (like web flash or Chrome's InfoBar)
 
-    You shouldn't reuse InfoBar object after close(). It is automatically
-    deleted.
-
     Layout::
 
         |widgets ...                |right widgets ...|x|
     """
+    finished = pyqtSignal(int)  # mimic QDialog
     linkActivated = pyqtSignal(unicode)
 
     # type of InfoBar (the number denotes its priority)
@@ -871,8 +975,6 @@ class InfoBar(QFrame):
     def __init__(self, parent=None):
         super(InfoBar, self).__init__(parent, frameShape=QFrame.StyledPanel,
                                       frameShadow=QFrame.Plain)
-        self.setAttribute(Qt.WA_DeleteOnClose)
-
         self.setAutoFillBackground(True)
         p = self.palette()
         p.setColor(QPalette.Window, QColor(self._colormap[self.infobartype]))
@@ -894,16 +996,29 @@ class InfoBar(QFrame):
     def addRightWidget(self, w):
         self.layout().insertWidget(self.layout().count() - 1, w)
 
+    def closeEvent(self, event):
+        if self.isVisible():
+            self.finished.emit(0)
+        super(InfoBar, self).closeEvent(event)
+
     def keyPressEvent(self, event):
         if event.key() == Qt.Key_Escape:
             self.close()
         super(InfoBar, self).keyPressEvent(event)
 
+    def heightForWidth(self, width):
+        # loosely based on the internal strategy of QBoxLayout
+        if self.layout().hasHeightForWidth():
+            return super(InfoBar, self).heightForWidth(width)
+        else:
+            return self.sizeHint().height()
+
 class StatusInfoBar(InfoBar):
     """Show status message"""
     def __init__(self, message, parent=None):
         super(StatusInfoBar, self).__init__(parent)
-        self._msglabel = QLabel(message, self, wordWrap=True,
+        self._msglabel = QLabel(message, self,
+                                wordWrap=True,
                                 textInteractionFlags=Qt.TextSelectableByMouse \
                                 | Qt.LinksAccessibleByMouse)
         self._msglabel.linkActivated.connect(self.linkActivated)
@@ -916,7 +1031,8 @@ class CommandErrorInfoBar(InfoBar):
     def __init__(self, message, parent=None):
         super(CommandErrorInfoBar, self).__init__(parent)
 
-        self._msglabel = QLabel(message, self, wordWrap=True,
+        self._msglabel = QLabel(message, self,
+                                wordWrap=True,
                                 textInteractionFlags=Qt.TextSelectableByMouse \
                                 | Qt.LinksAccessibleByMouse)
         self._msglabel.linkActivated.connect(self.linkActivated)
@@ -956,17 +1072,21 @@ class ConfirmInfoBar(InfoBar):
 
     def closeEvent(self, event):
         if self.isVisible():
+            self.finished.emit(1)
             self.rejected.emit()
+            self.hide()  # avoid double emission of finished signal
         super(ConfirmInfoBar, self).closeEvent(event)
 
     @pyqtSlot()
     def _accept(self):
+        self.finished.emit(0)
         self.accepted.emit()
         self.hide()
         self.close()
 
     @pyqtSlot()
     def _reject(self):
+        self.finished.emit(1)
         self.rejected.emit()
         self.hide()
         self.close()
@@ -1016,6 +1136,161 @@ class WidgetGroups(object):
     def set_enable(self, *args, **kargs):
         self.set_prop('setEnabled', *args, **kargs)
 
+class DialogKeeper(QObject):
+    """Manage non-blocking dialogs identified by creation parameters
+
+    Example "open single dialog per type":
+
+    >>> mainwin = QWidget()
+    >>> dialogs = DialogKeeper(lambda self, cls: cls(self), parent=mainwin)
+    >>> dlg1 = dialogs.open(QDialog)
+    >>> dlg1.parent() is mainwin
+    True
+    >>> dlg2 = dialogs.open(QDialog)
+    >>> dlg1 is dlg2
+    True
+    >>> dialogs.count()
+    1
+
+    closed dialog will be disowned:
+
+    >>> dlg1.reject()
+    >>> dlg1.parent() is None
+    True
+    >>> dialogs.count()
+    0
+
+    and recreates as necessary:
+
+    >>> dlg3 = dialogs.open(QDialog)
+    >>> dlg1 is dlg3
+    False
+
+    creates new dialog of the same type:
+
+    >>> dlg4 = dialogs.openNew(QDialog)
+    >>> dlg3 is dlg4
+    False
+    >>> dialogs.count()
+    2
+
+    and the last dialog is preferred:
+
+    >>> dialogs.open(QDialog) is dlg4
+    True
+    >>> dlg4.reject()
+    >>> dialogs.count()
+    1
+    >>> dialogs.open(QDialog) is dlg3
+    True
+
+    The following example is not recommended because it creates reference
+    cycles and makes hard to garbage-collect::
+
+        self._dialogs = DialogKeeper(self._createDialog)
+        self._dialogs = DialogKeeper(lambda *args: Foo(self))
+
+    When to delete reference:
+
+    If a dialog is not referenced, and if accept(), reject() and done() are
+    not overridden, it could be garbage-collected during finished signal and
+    lead to hard crash::
+
+        #0  isSignalConnected (signal_index=..., this=0x0)
+        #1  QMetaObject::activate (...)
+        #2  ... in sipQDialog::done (this=0x1d655b0, ...)
+        #3  ... in sipQDialog::reject (this=0x1d655b0)
+        #4  ... in QDialog::closeEvent (this=this@entry=0x1d655b0, ...)
+
+    To avoid crash, a finished dialog is referenced explicitly by DialogKeeper
+    until next event processing.
+
+    >>> dialogs = DialogKeeper(QDialog)
+    >>> dlgref = weakref.ref(dialogs.open())
+    >>> dialogs.count()
+    1
+    >>> esckeyev = QKeyEvent(QEvent.KeyPress, Qt.Key_Escape, Qt.NoModifier)
+    >>> QApplication.postEvent(dlgref(), esckeyev)  # close without incref
+    >>> QApplication.processEvents()  # should not crash
+    >>> dialogs.count()
+    0
+    >>> dlgref() is None  # nobody should have reference
+    True
+    """
+
+    def __init__(self, createdlg, genkey=None, parent=None):
+        super(DialogKeeper, self).__init__(parent)
+        self._createdlg = createdlg
+        self._genkey = genkey or DialogKeeper._defaultgenkey
+        self._keytodlgs = {}  # key: [dlg, ...]
+        self._dlgtokey = {}   # dlg: key
+
+        self._garbagedlgs = []
+        self._emptygarbagelater = QTimer(self, interval=0, singleShot=True)
+        self._emptygarbagelater.timeout.connect(self._emptygarbage)
+
+    def open(self, *args, **kwargs):
+        """Create new dialog or reactivate existing dialog"""
+        dlg = self._preparedlg(self._genkey(self.parent(), *args, **kwargs),
+                               args, kwargs)
+        dlg.show()
+        dlg.raise_()
+        dlg.activateWindow()
+        return dlg
+
+    def openNew(self, *args, **kwargs):
+        """Create new dialog even if there exists the specified one"""
+        dlg = self._populatedlg(self._genkey(self.parent(), *args, **kwargs),
+                                args, kwargs)
+        dlg.show()
+        dlg.raise_()
+        dlg.activateWindow()
+        return dlg
+
+    def _preparedlg(self, key, args, kwargs):
+        if key in self._keytodlgs:
+            assert len(self._keytodlgs[key]) > 0
+            return self._keytodlgs[key][-1]  # prefer latest
+        else:
+            return self._populatedlg(key, args, kwargs)
+
+    def _populatedlg(self, key, args, kwargs):
+        dlg = self._createdlg(self.parent(), *args, **kwargs)
+        if key not in self._keytodlgs:
+            self._keytodlgs[key] = []
+        self._keytodlgs[key].append(dlg)
+        self._dlgtokey[dlg] = key
+        dlg.finished.connect(self._forgetdlg)
+        return dlg
+
+    #@pyqtSlot()
+    def _forgetdlg(self):
+        dlg = self.sender()
+        dlg.finished.disconnect(self._forgetdlg)
+        if dlg.parent() is self.parent():
+            dlg.setParent(None)  # assist gc
+        key = self._dlgtokey.pop(dlg)
+        self._keytodlgs[key].remove(dlg)
+        if not self._keytodlgs[key]:
+            del self._keytodlgs[key]
+
+        # avoid deletion inside finished signal
+        self._garbagedlgs.append(dlg)
+        self._emptygarbagelater.start()
+
+    @pyqtSlot()
+    def _emptygarbage(self):
+        del self._garbagedlgs[:]
+
+    def count(self):
+        assert len(self._dlgtokey) == sum(len(dlgs) for dlgs
+                                          in self._keytodlgs.itervalues())
+        return len(self._dlgtokey)
+
+    @staticmethod
+    def _defaultgenkey(_parent, *args, **_kwargs):
+        return args
+
 class TaskWidget(object):
     def canswitch(self):
         """Return True if the widget allows to switch away from it"""
@@ -1082,6 +1357,17 @@ class Spacer(QWidget):
     def sizeHint(self):
         return QSize(self.width, self.height)
 
+def _configuredusername(ui):
+    # need to check the existence before calling ui.username(); otherwise it
+    # may fall back to the system default.
+    if (not os.environ.get('HGUSER') and not ui.config('ui', 'username')
+        and not os.environ.get('EMAIL')):
+        return None
+    try:
+        return ui.username()
+    except error.Abort:
+        return None
+
 def getCurrentUsername(widget, repo, opts=None):
     if opts:
         # 1. Override has highest priority
@@ -1090,10 +1376,9 @@ def getCurrentUsername(widget, repo, opts=None):
             return user
 
     # 2. Read from repository
-    try:
-        return repo.ui.username()
-    except error.Abort:
-        pass
+    user = _configuredusername(repo.ui)
+    if user:
+        return user
 
     # 3. Get a username from the user
     QMessageBox.information(widget, _('Please enter a username'),
@@ -1103,10 +1388,7 @@ def getCurrentUsername(widget, repo, opts=None):
     dlg = SettingsDialog(False, focus='ui.username')
     dlg.exec_()
     repo.invalidateui()
-    try:
-        return repo.ui.username()
-    except error.Abort:
-        return None
+    return _configuredusername(repo.ui)
 
 class _EncodingSafeInputDialog(QInputDialog):
     def accept(self):
@@ -1174,21 +1456,21 @@ class PaletteSwitcher(object):
     enablefilterpalette() method.
     """
     def __init__(self, targetwidget):
-        self._targetwidget = targetwidget
+        self._targetwref = weakref.ref(targetwidget)  # avoid circular ref
         self._defaultpalette = targetwidget.palette()
-        bgcolor = self._defaultpalette.color(QPalette.Base)
-        if bgcolor.black() <= 128:
-            # Light theme
+        if not isdarktheme(self._defaultpalette):
             filterbgcolor = QColor('#FFFFB7')
         else:
-            # Dark theme
             filterbgcolor = QColor('darkgrey')
         self._filterpalette = QPalette()
         self._filterpalette.setColor(QPalette.Base, filterbgcolor)
 
     def enablefilterpalette(self, enabled=False):
+        targetwidget = self._targetwref()
+        if not targetwidget:
+            return
         if enabled:
             pl = self._filterpalette
         else:
             pl = self._defaultpalette
-        self._targetwidget.setPalette(pl)
+        targetwidget.setPalette(pl)
