@@ -6,13 +6,11 @@
 # of the GNU General Public License, incorporated herein by reference.
 
 import cStringIO
-import os
 
-from mercurial import hg, util, patch, commands
+from mercurial import patch
 from hgext import record
 
 from tortoisehg.util import hglib
-from tortoisehg.util.patchctx import patchctx
 from tortoisehg.hgqt.i18n import _
 from tortoisehg.hgqt import qtlib, qscilib, lexers
 
@@ -23,7 +21,7 @@ from PyQt4 import Qsci
 qsci = Qsci.QsciScintilla
 
 class RejectsDialog(QDialog):
-    def __init__(self, path, parent):
+    def __init__(self, ui, path, parent=None):
         super(RejectsDialog, self).__init__(parent)
         self.setWindowTitle(_('Merge rejected patch chunks into %s') %
                             hglib.tounicode(path))
@@ -35,8 +33,6 @@ class RejectsDialog(QDialog):
         editor.setBraceMatching(qsci.SloppyBraceMatch)
         editor.setFolding(qsci.BoxedTreeFoldStyle)
         editor.installEventFilter(qscilib.KeyPressInterceptor(self))
-        editor.setContextMenuPolicy(Qt.CustomContextMenu)
-        editor.customContextMenuRequested.connect(self.menuRequested)
         self.baseLineColor = editor.markerDefine(qsci.Background, -1)
         editor.setMarkerBackgroundColor(QColor('lightblue'), self.baseLineColor)
         self.layout().addWidget(editor, 3)
@@ -109,8 +105,10 @@ class RejectsDialog(QDialog):
         f.seek(0)
         editor.read(f)
         editor.setModified(False)
-        lexer = lexers.get_lexer(path, earlybytes, self)
+        lexer = lexers.getlexer(ui, path, earlybytes, self)
         editor.setLexer(lexer)
+        if lexer is None:
+            editor.setFont(qtlib.getfont('fontlog').font())
         editor.setMarginLineNumbers(1, True)
         editor.setMarginWidth(1, str(editor.lines())+'X')
 
@@ -135,10 +133,6 @@ class RejectsDialog(QDialog):
         self.unresolved.setDisabled(True)
         QTimer.singleShot(0, lambda: self.chunklist.setCurrentRow(0))
 
-    def menuRequested(self, point):
-        point = self.editor.viewport().mapToGlobal(point)
-        return self.editor.createStandardContextMenu().exec_(point)
-
     def updateChunkList(self):
         self.updating = True
         self.chunklist.clear()
@@ -214,8 +208,11 @@ class RejectsDialog(QDialog):
 
         if acceptresolution:
             f = QFile(hglib.tounicode(self.path))
-            f.open(QIODevice.WriteOnly)
-            self.editor.write(f)
+            saved = f.open(QIODevice.WriteOnly) and self.editor.write(f)
+            if not saved:
+                qtlib.ErrorMsgBox(_('Unable to save file'),
+                                  f.errorString(), parent=self)
+                return
             self.saveSettings()
             super(RejectsDialog, self).accept()
 
@@ -233,8 +230,6 @@ class RejectBrowser(qscilib.Scintilla):
         self.setUtf8(True)
 
         self.installEventFilter(qscilib.KeyPressInterceptor(self))
-        self.setContextMenuPolicy(Qt.CustomContextMenu)
-        self.customContextMenuRequested.connect(self.menuRequested)
         self.setCaretLineVisible(False)
 
         self.setMarginType(1, qsci.SymbolMargin)
@@ -250,13 +245,9 @@ class RejectBrowser(qscilib.Scintilla):
         mask = (1 << self.addedMark) | (1 << self.removedMark) | \
                (1 << self.addedColor) | (1 << self.removedColor)
         self.setMarginMarkerMask(1, mask)
-        lexer = lexers.get_diff_lexer(self)
+        lexer = lexers.difflexer(self)
         self.setLexer(lexer)
 
-    def menuRequested(self, point):
-        point = self.viewport().mapToGlobal(point)
-        return self.createStandardContextMenu().exec_(point)
-
     def showChunk(self, lines):
         utext = []
         added = []
@@ -275,14 +266,3 @@ class RejectBrowser(qscilib.Scintilla):
         for i in removed:
             self.markerAdd(i, self.removedMark)
             self.markerAdd(i, self.removedColor)
-
-def run(ui, *pats, **opts):
-    if len(pats) != 1:
-        qtlib.ErrorMsgBox(_('Filename required'),
-                          _('You must provide the path to a file'))
-        import sys; sys.exit()
-    path = pats[0]
-    if path.endswith('.rej'):
-        path = path[:-4]
-    dlg = RejectsDialog(path, None)
-    return dlg
